`timescale 1ns / 1ps
////////////////////////////////////////////////////////////////////////////////
// Engineer:       Michael Siegenthaler
//
// Design Name:    EPP I/O Slave
// Module Name:    eppioslave
// Revision:       July 17, 2005
// 
////////////////////////////////////////////////////////////////////////////////
module eppioslave(mclk, epp_ctrl_write, epp_ctrl_dstb, epp_ctrl_rst, epp_ctrl_astb, epp_sta_wait, epp_data_bus,
                  ram_clk, ram_addr, ram_en, ram_we, ram_di, ram_do);

  // Inputs/outputs
  input mclk;
  input epp_ctrl_write;                                                                                       
  input epp_ctrl_dstb;                                                                                       
  input epp_ctrl_rst;                                                                                       
  input epp_ctrl_astb;                                                                                       
  output epp_sta_wait;
  inout [7:0] epp_data_bus;
  output ram_clk;
  output [7:0] ram_addr;
  output ram_en;
  output ram_we;
  output [7:0] ram_di;
  input [7:0] ram_do;

  // External nodes
  wire mclk;
  wire epp_ctrl_write;
  wire epp_ctrl_dstb;
  wire epp_ctrl_rst;
  wire epp_ctrl_astb;
  wire epp_sta_wait;
  wire [7:0] epp_data_bus;
  wire ram_clk;
  reg [7:0] ram_addr;
  wire ram_en;
  reg ram_we;
  wire [7:0] ram_di;
  wire [7:0] ram_do;

  // Internal nodes
  reg [7:0] data_bus_out;
  reg [7:0] data_bus_in;
  reg output_enable;
  reg ctrl_write;
  reg ctrl_dstb;
  reg ctrl_rst;
  reg ctrl_astb;
  reg sta_wait;

  assign epp_sta_wait = sta_wait;
  assign ram_en = 1'b1;

  reg [3:0] state;

  // Clock divider
  wire clk;
  reg [3:0] clkcount;
  assign clk = clkcount[3];
  assign ram_clk = clkcount[3];
  always @(negedge mclk) begin // use negedge to eliminate timing violation in dual-port ram
    clkcount <= clkcount+1;
  end

  // Tristate buffer for data_bus
  assign epp_data_bus = output_enable ? data_bus_out : 8'bz;
  assign ram_di = data_bus_in;

  // Register inputs
  always @(posedge mclk) begin
    data_bus_in <= epp_data_bus;
    ctrl_write <= epp_ctrl_write;
    ctrl_dstb <= epp_ctrl_dstb;
    ctrl_rst <= epp_ctrl_rst;
    ctrl_astb <= epp_ctrl_astb;
  end

  // State machine
  always @(posedge clk) begin

    // All read/write labels are from the perspective of the host computer

    // Reset state
    if (ctrl_rst == 1'b0) begin
      state <= 4'd1; // Go to branch state as soon as ctrl_rst=1
      sta_wait <= 1'b0;
      output_enable <= 1'b0;
      ram_we <= 1'b0;
    end

    // Branch
    // Wait for an address or data strobe and transition to the appropriate state
    else if (state == 4'd1) begin
      if      (ctrl_dstb==1'b0 && ctrl_write==1'b0) state <= 4'd2;
      else if (ctrl_dstb==1'b0 && ctrl_write==1'b1) state <= 4'd4;
      else if (ctrl_astb==1'b0 && ctrl_write==1'b0) state <= 4'd6;
      else if (ctrl_astb==1'b0 && ctrl_write==1'b1) state <= 4'd8;
      sta_wait <= 1'b0;
    end

    // Data write cycle, step 1
    else if (state == 4'd2) begin
      if (ctrl_dstb==1'b1) state <= 4'd3;
      sta_wait <= 1'b1;
      ram_we <= 1'b1;
    end

    // Data write cycle, step 2
    else if (state == 4'd3) begin
      state <= 4'd1;
      sta_wait <= 1'b0;
      ram_we <= 1'b0;
    end

    // Address write cycle, step 1
    else if (state == 4'd6) begin
      if (ctrl_astb==1'b1) state <= 4'd7;
      sta_wait <= 1'b1;
    end

    // Address write cycle, step 2
    else if (state == 4'd7) begin
      state <= 4'd1;
      sta_wait <= 1'b0;
      ram_addr <= data_bus_in; // Latch written address
    end

    // Data read cycle, step 1
    else if (state == 4'd4) begin
      if (ctrl_dstb==1'b1) state <= 4'd5;
      sta_wait <= 1'b1;
      output_enable <= 1'b1;
      data_bus_out <= ram_do;
    end

    // Data read cycle, step 2
    else if (state == 4'd5) begin
      state <= 4'd1;
      output_enable <= 1'b0;
    end

    // Address read cycle, step 1
    else if (state == 4'd8) begin
      if (ctrl_astb==1'b1) state <= 4'd9;
      sta_wait <= 1'b1;
      output_enable <= 1'b1;
      data_bus_out <= ram_addr;
    end

    // Address read cycle, step 2
    else if (state == 4'd9) begin
      state <= 4'd1;
      output_enable <= 1'b0;
    end

    else begin
      // we should never get here, but just in case...
      state <= 4'd1;
      sta_wait <= 1'b0;
      output_enable <= 1'b0;
      ram_we <= 1'b0;
    end

  end

endmodule
